import enum
from numbers import Real
from typing import Any, Callable, Iterable, List, Optional, Tuple

from dataclassy import dataclass
from dataclassy.dataclass import Internal
from dataclassy.functions import replace

from wordmaze.utils.dataclasses import DataClassSequence
from wordmaze.utils.sequences import MutableSequence


@dataclass(iter=True, kwargs=True)
class Box:
    x1: Real = None
    x2: Real = None
    y1: Real = None
    y2: Real = None

    def __post_init__(
            self,
            height: Optional[Real] = None,
            width: Optional[Real] = None
    ) -> None:
        if (self.x1, self.x2, width).count(None) != 1:
            raise ValueError(
                'exactly one of *x1*, *x2*, *width* must be *None* or omitted'
            )
        elif width is not None and width < 0:
            raise ValueError('*width* must be *None* (or omitted) or >=0')

        if (self.y1, self.y2, height).count(None) != 1:
            raise ValueError(
                'exactly one of *y1*, *y2*, *height* must be *None* (or omitted)'
            )
        elif height is not None and height < 0:
            raise ValueError('*height* must be *None* (or omitted) or >=0')

        if self.x1 is None:
            self.x1 = self.x2 - width
        elif self.x2 is None:
            self.x2 = self.x1 + width
        else:
            if self.x1 > self.x2:
                self.x2, self.x1 = self.x1, self.x2
            width = self.x2 - self.x1

        if self.y1 is None:
            self.y1 = self.y2 - height
        elif self.y2 is None:
            self.y2 = self.y1 + height
        else:
            if self.y1 > self.y2:
                self.y2, self.y1 = self.y1, self.y2
            height = self.y2 - self.y1

        self._height: Real = height
        self._width: Real = width

    @property
    def height(self) -> Real:
        return self._height

    @property
    def width(self) -> Real:
        return self._width


class TextBox(Box):
    text: str
    confidence: Real


@dataclass(iter=True)
class Shape:
    height: Real
    width: Real


class Origin(enum.Enum):
    TOP_LEFT = enum.auto()
    BOTTOM_LEFT = enum.auto()


class Page(DataClassSequence[TextBox]):
    def __init__(
            self,
            shape: Shape,
            entries: Iterable[TextBox] = (),
            origin: Origin = Origin.TOP_LEFT
    ) -> None:
        super().__init__(entries)
        self.shape: Shape = shape
        self.origin: Origin = origin

    def map(
            self,
            *mapper: Callable[[TextBox], TextBox],
            **field_mappers: Callable[[Any], Any]
    ) -> 'Page':
        return Page(
            shape=self.shape,
            origin=self.origin,
            entries=super().map(*mapper, **field_mappers)
        )

    def filter(
            self,
            *pred: Callable[[TextBox], bool],
            **field_preds: Callable[[Any], bool]
    ) -> 'Page':
        return Page(
            shape=self.shape,
            origin=self.origin,
            entries=super().filter(*pred, **field_preds)
        )

    def rebase(self, origin: Origin) -> 'Page':
        if origin is self.origin:
            return self
        elif (
            (origin is Origin.BOTTOM_LEFT and self.origin is Origin.TOP_LEFT)
            or (origin is Origin.TOP_LEFT and self.origin is Origin.BOTTOM_LEFT)
        ):
            def rebaser(textbox: TextBox) -> TextBox:
                return replace(
                    textbox,
                    y1=self.shape.height - textbox.y2,
                    y2=self.shape.height - textbox.y1
                )
        else:
            raise NotImplementedError(
                'unsupported rebase operation:'
                f' from {self.origin} to {origin}.'
            )

        rebased = self.map(rebaser)
        rebased.origin = origin
        return rebased


class WordMaze(MutableSequence[Page]):
    def __init__(self, pages: Iterable[Page] = ()) -> None:
        super().__init__(pages)

    @property
    def shapes(self) -> Tuple[Shape, ...]:
        return tuple(page.shape for page in self)

    def tuples(self) -> Iterable[tuple]:
        return (
            (number,) + tpl
            for number, page in enumerate(self)
            for tpl in page.tuples()
        )

    def dicts(self) -> Iterable[dict]:
        return (
            dict(dct, page=number)
            for number, page in enumerate(self)
            for dct in page.dicts()
        )

    def map(
            self,
            *mapper: Callable[[TextBox], TextBox],
            **field_mappers: Callable[[Any], Any]
    ) -> 'WordMaze':
        return WordMaze(page.map(*mapper, **field_mappers) for page in self)

    def filter(
            self,
            *pred: Callable[[TextBox], bool],
            **field_preds: Callable[[Any], bool]
    ) -> 'WordMaze':
        return WordMaze(page.filter(*pred, **field_preds) for page in self)



from typing import Iterable, List
from typing import MutableSequence as BaseSequence
from typing import TypeVar

_Entry = TypeVar('_Entry')


class MutableSequence(BaseSequence[_Entry]):
    def __init__(self, entries: Iterable[_Entry] = ()) -> None:
        self.__entries: List[_Entry] = list(entries)

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.__entries})'

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, index: int) -> _Entry:
        return self.__entries[index]

    def __setitem__(self, index: int, entry: _Entry) -> None:
        self.__entries[index] = entry

    def __delitem__(self, index: int) -> None:
        del self.__entries[index]

    def __len__(self) -> int:
        return len(self.__entries)

    def insert(self, index: int, entry: _Entry) -> None:
        self.__entries.insert(index, entry)



import textwrap
from functools import partial
from typing import Any, Callable, Dict, Generic, Iterable, List, TypeVar

import funcy
from dataclassy import DataClass, asdict, astuple, values
from dataclassy.functions import is_dataclass_instance, replace

from wordmaze.utils.sequences import MutableSequence

_DataClass = TypeVar('_DataClass', bound=DataClass)


def as_dict(
        obj: DataClass,
        flatten: bool = False
) -> Dict[str, Any]:
    if flatten:
        return funcy.join(
            (
                as_dict(field, flatten=flatten)
                if is_dataclass_instance(field)
                else {field_name: field}
            )
            for field_name, field in values(obj).items()
        )
    else:
        return asdict(obj)


def as_tuple(
        obj: DataClass,
        flatten: bool = False
) -> tuple:
    if flatten:
        return funcy.join(
            (
                as_tuple(field, flatten=flatten)
                if is_dataclass_instance(field)
                else (field,)
            )
            for field in values(obj).values()
        )
    else:
        return astuple(obj)


def field_mapper(
        *mapper: Callable[[_DataClass], _DataClass],
        **field_mappers: Callable[[Any], Any]
) -> Callable[[_DataClass], _DataClass]:
    if mapper and field_mappers:
        raise TypeError(textwrap.dedent(
            '''
            *field_mapper* accepts either a mapper or keyworded mappers. For instance:
                field_mapper(lambda textbox: process_textbox(textbox))
                field_mapper(
                    x1=lambda x1: x1 + 5,
                    x2=lambda x2: x2**2,
                    text=lambda text: text.upper()
                )
            '''
        ))

    if field_mappers:
        def _mapper(obj: _DataClass) -> _DataClass:
            obj_dict = as_dict(obj)

            keys_diff = frozenset(field_mappers).difference(obj_dict)
            if keys_diff:
                raise ValueError(
                    f'fields {list(keys_diff)} do not exist in object'
                    f' {obj} from the class {type(obj)}.'
                )

            obj_dict = funcy.project(obj_dict, field_mappers)
            changes = {
                field_name: _field_mapper(obj_dict[field_name])
                for field_name, _field_mapper in field_mappers.items()
            }

            return replace(
                obj,
                **changes
            )
    else:
        _mapper = mapper[0]

    return _mapper


def field_pred(
        *pred: Callable[[_DataClass], bool],
        **field_preds: Callable[[Any], bool]
) -> Callable[[_DataClass], bool]:
    if pred and field_preds:
        raise TypeError(textwrap.dedent(
        '''
        invalid call. *field_pred* accepts either a pred or keyworded preds. For instance:
        field_pred(lambda textbox: process_textbox(textbox))
        field_pred(
            x1=lambda x1: x1 >= 0,
            x2=lambda x2: (x2 % 2)==0,
            text=lambda text: text.isdigit()
        )
        '''))

    if field_preds:
        def _pred(obj: _DataClass) -> bool:
            obj_dict = as_dict(obj)

            keys_diff = frozenset(field_preds).difference(obj_dict)
            if keys_diff:
                raise ValueError(
                    f'fields {list(keys_diff)} do not exist in object'
                    f' {obj} from the class {type(obj)}.'
                )

            obj_dict = funcy.project(obj_dict, field_preds)
            preds = (
                _field_pred(obj_dict[field_name])
                for field_name, _field_pred in field_preds.items()
            )

            return all(preds)
    else:
        _pred = pred[0]

    return _pred


class DataClassSequence(MutableSequence[_DataClass]):
    def __init__(self, entries: Iterable[_DataClass] = ()) -> None:
        super().__init__(entries)

    def tuples(self) -> Iterable[tuple]:
        return map(
            partial(as_tuple, flatten=True),
            self
        )

    def dicts(self) -> Iterable[dict]:
        return map(
            partial(as_dict, flatten=True),
            self
        )

    def map(
            self,
            *mapper: Callable[[_DataClass], _DataClass],
            **field_mappers: Callable[[Any], Any]
    ) -> Iterable[_DataClass]:
        _mapper = field_mapper(*mapper, **field_mappers)
        return map(_mapper, self)

    def filter(
            self,
            *pred: Callable[[_DataClass], bool],
            **field_preds: Callable[[Any], bool]
    ) -> Iterable[_DataClass]:
        _pred = field_pred(*pred, **field_preds)
        return filter(_pred, self)



#!/usr/bin/env python

from pathlib import Path

from setuptools import find_packages, setup

here = Path(__file__).resolve().parent

VERSION = (here / 'VERSION').read_text().strip()

setup(
    name='wordmaze',
    version=VERSION,

    description="Words and textboxes made amazing",
    long_description=Path(here, 'README.md').read_text(),
    long_description_content_type="text/markdown",

    license='MIT',
    author='elint-tech',
    author_email='contato@elint.com.br',

    url='https://github.com/elint-tech/wordmaze/',
    download_url=f'https://github.com/elint-tech/wordmaze/dist/wordmaze-{VERSION}.tar.gz',

    install_requires=Path(here, 'requirements.txt').read_text().splitlines(),
    packages=find_packages(),

    keywords=[
        'wordmaze',
        'ocr',
        'pdf',
        'word',
        'words',
        'text',
        'textbox',
        'textboxes',
        'text-mining',
        'text-mining-analysis'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ]
)
